
#include "libpqtypes-int.h"

int
PQgetf(const PGresult *res, int tup_num, const char *fieldspec, ...)
{
	int r;
	PGtypeHandler *h;
	PGtypeArgs args;
	int typpos = 0;
	char *s;
	int flags;
	Oid ftype;
	char schema[PQT_MAXIDLEN + 1];
	char typname[PQT_MAXIDLEN + 1];
	PGtypeData *resData;

	if(!res)
	{
		PQseterror("Invalid argument - PGresult pointer cannot be NULL");
		return 0;
	}

	if(!(resData = PQresultHookData(res, PQT_OBJHOOKSNAME)))
	{
		PQseterror("PGresult is missing libpqtypes hook data");
		return 0;
	}

	va_start(args.ap, fieldspec);

	while(fieldspec && *fieldspec)
	{
		int specid = *fieldspec++;

		if(specid != '%' && specid != '#')
			continue;

		typpos++;

		if(!(s = pqt_parsetype(
				fieldspec,
				schema,  /* out buffer */
				typname, /* out buffer */
				&flags,
				typpos)))
		{
			va_end(args.ap);
			return 0;
		}

		if(flags & TYPFLAG_INVALID)
		{
			va_end(args.ap);
			return 0;
		}

		fieldspec = (char *)s;

		/* find the handler */
		if(!(h = pqt_gethandler(
			resData->typhandlers,
			resData->typhcnt,
			schema,
			typname)))
		{
			PQseterror("Uknown typname '%s' (position %d)",
				typname, typpos);
			va_end(args.ap);
			return 0;
		}

		if(!h->typget)
		{
			h = pqt_gethandlerbyid(
				resData->typhandlers, resData->typhcnt, h->base_id);

			if(!h || !h->typget)
			{
				PQseterror(
					"typname '%s' doesn't support get operations (position %d)",
					h->typname, typpos);
				va_end(args.ap);
				return 0;
			}
		}

		if(specid == '#')
			args.get.field_num = PQfnumber(res, va_arg(args.ap, const char *));
		else
			args.get.field_num = va_arg(args.ap, int);

		/* simplify life for handlers by checking getvalue's return here. */
		if(args.get.field_num < 0 || !PQgetvalue(res, tup_num, args.get.field_num))
		{
			PQseterror(
				"invalid tup_num[%d].field_num[%d] (position %d)",
				tup_num, args.get.field_num, typpos);
			va_end(args.ap);
			return 0;
		}

		ftype = PQftype(res, args.get.field_num);
		if(((flags & TYPFLAG_ARRAY) && ftype != h->typoid_array) ||
			 (!(flags & TYPFLAG_ARRAY) && ftype != h->typoid))
		{
			Oid oid = (flags & TYPFLAG_ARRAY) ? h->typoid_array : h->typoid;
			PQseterror(
				"Trying to get type %u (%s) but server returned %u (position %d)",
				oid, h->typname, ftype, typpos);
			va_end(args.ap);
			return 0;
		}

		args.is_put       = 0;
		args.get.result   = (PGresult *)res;
		args.format       = PQfformat(res, args.get.field_num);
		args.fmtinfo      = &resData->fmtinfo;
		args.get.tup_num  = tup_num;
		args.is_ptr       = (flags & TYPFLAG_POINTER) ? 1 : 0;
		args.typpos       = typpos;
		args.typhandler   = h;
		args.errorf       = pqt_argserrorf;
		args.super        = pqt_argssuper;

		if(flags & TYPFLAG_ARRAY)
			r = pqt_get_array(&args);
		else
			r = h->typget(&args);

		if(r == -1)
		{
			va_end(args.ap);
			return 0;
		}
	}

	va_end(args.ap);
	return 1;
}

/* --------------------------------
 * Exec and Send functions
 */

/*
 * Each param value requires an oid, ptr and 2 ints (oid, value, length, format).
 * This makes the maximum size of a param 20 bytes. A stack size of 4k
 * would allow for 204 query params.	If more params are needed,
 * heap memory is used. Needing more than 204 param values in a query
 * is very rare, although possible.
 */
#define PARAM_STACKSIZE 4096

#define BUILD_ARRAYS(rettype) \
	rettype r; \
	char *buf   = NULL; \
	Oid *oids   = NULL; \
	char **vals = NULL; \
	int *lens   = NULL; \
	int *fmts   = NULL; \
	int vcnt    = 0; \
	char stackbuffer[PARAM_STACKSIZE]; \
\
	if(!conn) \
		return (rettype)(0); \
\
	if(param) \
	{ \
		buf = stackbuffer; \
		if(!buildarrays(param, &buf, &oids, &vals, &lens, &fmts)) \
			return (rettype)(0); \
		vcnt = param->vcnt; \
	}

#define RETURN_RESULT \
	if(param) \
	{ \
		if(buf && buf != stackbuffer) \
			free(buf); \
	} \
	return r

/*
 * Assigns param values to param arrays, for use with postgres
 * parameter API. 'buf' is expected to be PARAM_STACKSIZE bytes.  If more
 * memory is required, memory is allocated and assigned to *buf, which
 * must be freed by caller.  To determine if *buf was allocated, compare
 * its address to the initially provided stack address.
 * Returns 1 on success and 0 on error.
 */
static int
buildarrays(PGparam *param, char **buf, Oid **oids,
	char ***vals, int **lens, int **fmts);

PGresult *
PQparamExec(PGconn *conn, PGparam *param, const char *command,
	int resultFormat)
{
	BUILD_ARRAYS(PGresult *);

	r = PQexecParams(
		conn,
		command,
		vcnt,
		oids,
		(const char *const * )vals,
		lens,
		fmts,
		resultFormat);

	RETURN_RESULT;
}

int
PQparamSendQuery(PGconn *conn, PGparam *param, const char *command,
	int resultFormat)
{
	BUILD_ARRAYS(int);

	r = PQsendQueryParams(
		conn,
		command,
		vcnt,
		oids,
		(const char *const * )vals,
		lens,
		fmts,
		resultFormat);

	RETURN_RESULT;
}

PGresult *
PQparamExecPrepared(PGconn *conn, PGparam *param, const char *stmtName,
	int resultFormat)
{
	BUILD_ARRAYS(PGresult *);

	r = PQexecPrepared(
		conn,
		stmtName,
		vcnt,
		(const char *const * )vals,
		lens,
		fmts,
		resultFormat);

	RETURN_RESULT;
}

int
PQparamSendQueryPrepared(PGconn *conn, PGparam *param, const char *stmtName,
	int resultFormat)
{
	BUILD_ARRAYS(int);

	r = PQsendQueryPrepared(
		conn,
		stmtName,
		vcnt,
		(const char *const * )vals,
		lens,
		fmts,
		resultFormat);

	RETURN_RESULT;
}

static int
buildarrays(PGparam *param, char **buf, Oid **oids,
	char ***vals, int **lens, int **fmts)
{
	int n;

	/* no params to assign */
	if(param->vcnt == 0)
		return 1;

	/* required memory size for the 4 param arrays */
	n = (int)((sizeof(void *) * param->vcnt) + /* values */
			((sizeof(int) * 2) * param->vcnt) +	   /* lengths and formats */
			(sizeof(Oid) * param->vcnt));				   /* oids */

	/* required memory is too large for stack buffer, get some heap */
	if(n > PARAM_STACKSIZE)
	{
		char *p;

		if(!(p = (char *)malloc(n)))
		{
			PQseterror(PQT_OUTOFMEMORY);
			return 0;
		}

		*buf = p;
	}

	/* give arrays memory from buffer, which could be stack or heap. */
	*vals = (char **)*buf;
	*lens = (int *)(*buf + (sizeof(void *) * param->vcnt));
	*fmts = (*lens) + param->vcnt;
	*oids = (Oid *)((*fmts) + param->vcnt);

	/* loop param values and assign value, length, format and oid to arrays. */
	for(n=0; n < param->vcnt; n++)
	{
		(*oids)[n] = param->vals[n].oid;
		(*vals)[n] = param->vals[n].data;
		(*lens)[n] = param->vals[n].datal;
		(*fmts)[n] = param->vals[n].format;
	}

	return 1;
}

