
#ifndef LIBPQTYPES_H
#define LIBPQTYPES_H

#include <libpq-fe.h>

#ifdef __cplusplus
extern "C" {
#endif

#if defined(_MSC_VER) || defined(__MINGW32__) || defined(__CYGWIN__)
#	define PQT_EXPORT __declspec(dllexport)
#else
#	define PQT_EXPORT extern
#endif

#define PQT_OBJHOOKSNAME "libpqtypes"

/*
 * PGparam encapsulates the building of parameterized arguments.  Params
 * are fully managed by this structure.
 */
typedef struct pg_param PGparam;

/* Useful connection-based info for type handlers. */
typedef struct
{
  int sversion;          /* server version */
	int pversion;          /* protocol version */
  char datestyle[48];    /* server DateStyle, like "SQL, MDY" */
  int integer_datetimes; /* When non-zero, server uses int64 timestamps */
} PGtypeFormatInfo;

typedef struct pg_typeargs PGtypeArgs;

/*
 * When used for puts, the return is the number of bytes
 * put or -1 for error.  When used for gets, the return is
 * 0 for success and -1 on error.
 */
typedef int (*PGtypeProc)(PGtypeArgs *args);

/* Record Attribute Description, its columns */
typedef struct
{
	Oid attoid;    /* oid of the composite column */
	int attlen;
	int atttypmod; /* type-specific data supplied at table creation time */
	char *attname; /* name of the composite column */
} PGrecordAttDesc;

/* Type handler for putf and getf functions. */
typedef struct pg_typhandler
{
	int id;
	char *typschema; /* Can be NULL if not provided (never NULL for builtins) */
	char *typname;
	int typlen;
	Oid typoid;
	Oid typoid_array;
	PGtypeProc typput;
	PGtypeProc typget;
	int base_id;

	/* Composites only, contains each attribute/column of a composite */
	int nattrs;
	PGrecordAttDesc *attDescs;

	/* For PQresets so we can recall PQregisterTypeHandler.  These are
	 * the user-provided arguments saved off.
	 */
	char *orig_typname;
	PGtypeProc orig_typput;
	PGtypeProc orig_typget;
} PGtypeHandler;

struct pg_typeargs
{
	int is_put;

	const PGtypeFormatInfo *fmtinfo;

	/* Indicates if a request for a direct pointer was made, %text*. */
	int is_ptr;

	/* When is_put is 1, set this to 1 for binary and 0 text format.  It
	 * defaults to binary.  When is_put is 0, this indicates the field
	 * type of get.field_num.
	 */
	int format;

	/* An argument list.  Arguments should be retrieved with va_arg. */
	va_list ap;

	/* The position of this typname within a specifier string, 1-based. */
	int typpos;

	/* Type handler for the specifier at typpos. */
	PGtypeHandler *typhandler;

	/*
	 * Report an error from within a handler.  This error message will show
	 * up in PQgeterror.
	 *
	 * This always returns -1 so one can report an error and return -1 from
	 * a handler in a single statement:
	 *
	 *   return args->errorf(args, "something failed: %s", strerror(errno));
	 *
	 * errorf always prepends a small header "schema.typname[pos:num] - msg".
	 * For example, if the above failed within the int4 handler and typpos
	 * was 5, the resulting error message would be:
	 *
	 *   pg_catalog.int4[pos:5] - something failed: Invalid argument
	 *
	 * errorf does not put any newlines in error message.
	 */
	int (*errorf)(PGtypeArgs *args, const char *format, ...);

	int (*super)(PGtypeArgs *args, ...);

	struct
	{
		PGparam *param;
		/*
		 * A buffer used to store the type's output format.  If more than
		 * 'outl' bytes are needed, see 'expandBuffer'.  Normally data is
		 * copied to the out buffer, but it can also be pointed elsewhere: like
		 * a const string or static memory.  When repointing the out buffer,
		 * DO NOT use 'expandBuffer'.  Never use realloc on this buffer.
		 */
		char *out;
		/* The size in bytes of the 'out' buffer. If expandBuffer is used,
		 * this will reflect the new buffer length.
		 */
		int outl;
		/*
		 * Expands the 'out' buffer to 'new_len'.  If new_len is less than or
		 * equal to the current length 'outl', the expand request is ignored.
		 * This behaves just like a realloc, existing data is copied to the new
		 * memory.  You should never use realloc on the out buffer.
		 * Returns -1 on error and 0 for success.
		 */
		int (*expandBuffer)(PGtypeArgs *args, int new_len);
		int _state; /* internal use only. */
	} put;

	struct
	{
		PGresult *result;
		int tup_num;
		int field_num;
	} get;
};

/* ----------------
 * Variable Length types
 * ----------------
 */

typedef char *PGtext;
typedef char *PGvarchar;
typedef char *PGbpchar;
typedef char *PGuuid;
typedef struct
{
  int len;
  char *data;
} PGbytea;

/* ----------------
 * Numeric types
 * ----------------
 */

typedef char PGchar;
typedef int PGbool;
typedef short PGint2;
typedef int PGint4;
typedef float PGfloat4;
typedef double PGfloat8;
typedef char *PGnumeric;

/* Defined by an end-user if the system is missing long long. */
#ifdef PQT_LONG_LONG
	typedef PQT_LONG_LONG PGint8;
  typedef PQT_LONG_LONG PGmoney;

/* MinGW and MSVC can use __int64 */
#elif defined(__MINGW32__) || defined(_MSC_VER)
  typedef __int64 PGint8;
	typedef __int64 PGmoney;

/* Cygwin and Unixes. */
#else
  typedef long long PGint8;
  typedef long long PGmoney;
#endif

/* ----------------
 * Geometric type structures
 * ----------------
 */

typedef struct
{
	double x;
	double y;
} PGpoint;

typedef struct
{
	PGpoint pts[2];
} PGlseg;

typedef struct
{
	PGpoint high;
	PGpoint low;
} PGbox;

typedef struct
{
	PGpoint center;
	double radius;
} PGcircle;

/* When used with PQgetf, 'pts' is only valid for the life of the result */
typedef struct
{
	int npts;
	int closed;
	PGpoint *pts;
} PGpath;

/* When used with PQgetf, 'pts' is only valid for the life of the result */
typedef struct
{
	int npts;
	PGpoint *pts;
} PGpolygon;

/* ----------------
 * Network type structures
 * ----------------
 */

/* This struct works with CIDR as well. */
typedef struct
{
	/* mask, ie. /24 */
	int mask;
	int is_cidr;

	/* Cast to: sockaddr, sockaddr_in, sockaddr_in6, sockaddr_stroage */
	int sa_len;
	char sa_buf[128];
} PGinet;

typedef struct
{
	int a;
	int b;
	int c;
	int d;
	int e;
	int f;
} PGmacaddr;

/* ----------------
 * Date & Time structures
 * ----------------
 */

typedef struct
{
	int years;
	int mons;
	int days;
	int hours;
	int mins;
	int secs;
	int usecs;
} PGinterval;

typedef struct
{
	/* When non-zero, the date is in the BC ERA. */
	int isbc;

	/*
	 * The BC or AD year, which is NOT adjusted by 1900 like
	 * the POSIX struct tm.  Years are always positive values,
	 * even BC years.  To distinguish between BC and AD years,
	 * use the isbc flag: (year 0 not used)
	 *   Ex. -1210 is represented as: isbc=1, year=1209
	 */
	int year;

	/* The number of months since January, in the range 0 to 11. */
	int mon;

	/* The day of the month, in the range 1 to 31. */
	int mday;

	/* The Julian day in the Gregorian calendar. */
	int jday;

	/* The number of days since January 1, in the range 0 to 365. */
	int yday;

	/* The number of days since Sunday, in the range 0 to 6. */
	int wday;
} PGdate;

typedef struct
{
	/* The number of hours past midnight, in the range 0 to 23. */
	int hour;

	/* The number of minutes after the hour, in the range 0 to 59. */
	int min;

	/* The number of seconds after the minute, in the range 0 to 59. */
	int sec;

	/* The number of microseconds after the second, in the
	 * range of 0 to 999999.
	 */
	int usec;

	/*
	 * When non-zero, this is a TIME WITH TIME ZONE.  Otherwise,
	 * it is a TIME WITHOUT TIME ZONE.
	 */
	int withtz;

	/* A value of 1 indicates daylight savings time.  A value of 0 indicates
	 * standard time.  A value of -1 means unknown or could not determine.
	 */
	int isdst;

	/* Seconds east of UTC. This value is not always available. It is
	 * set to 0 if it cannot be determined.
	 */
	int gmtoff;

	/* Timezone abbreviation: such as EST, GMT, PDT, etc. This value is
	 * not always available.  It is set to an empty string if it cannot be
	 * determined.
	 */
	char tzName[16];
} PGtime;

typedef struct
{
	/* The number seconds before or after midnight UTC of January 1, 1970,
	 * not counting leap seconds.
	 */
	PGint8 epoch;

	/* The date part of the timestamp. */
	PGdate date;

	/* The time part of the timestamp. */
	PGtime time;
} PGtimestamp;

/* ----------------
 * Array structures
 * ----------------
 */

#ifndef MAXDIM
#	define MAXDIM 6
#endif

typedef struct
{
	int ndims;
	int lbound[MAXDIM];
	int dims[MAXDIM];
	PGparam *param;
	PGresult *res;
} PGarray;

/* ----------------
 * Public API funcs
 * ----------------
 */

/* === in hooks.c === */

PQT_EXPORT PGobjectHooks *
PQtypesObjectHooks(void);

/* === in error.c === */

PQT_EXPORT const char *
PQgeterror(void);

PQT_EXPORT void
PQseterror(const char *format, ...);

/* === in handler.c === */

PQT_EXPORT int
PQregisterTypeHandler(PGconn *conn, const char *typname,
	PGtypeProc put, PGtypeProc get);

/* === in param.c === */

PQT_EXPORT PGparam *
PQparamCreate(const PGconn *conn);

PQT_EXPORT void
PQparamReset(PGparam *param);

PQT_EXPORT void
PQparamClear(PGparam *param);

PQT_EXPORT int
PQputf(PGparam *param, const char *paramSpec, ...);

PQT_EXPORT int
PQputvf(PGparam *param, char *stmtBuf, size_t stmtBufLen,
	const char *spec, va_list ap);

/* === in exec.c === */

PQT_EXPORT int
PQgetf(const PGresult *res, int tup_num, const char *fieldspec, ...);

PQT_EXPORT PGresult *
PQparamExec(PGconn *conn, PGparam *param,
	const char *command, int resultFormat);

PQT_EXPORT int
PQparamSendQuery(PGconn *conn, PGparam *param,
	const char *command, int resultFormat);

PQT_EXPORT PGresult *
PQparamExecPrepared(PGconn *conn, PGparam *param,
	const char *stmtName, int resultFormat);

PQT_EXPORT int
PQparamSendQueryPrepared(PGconn *conn, PGparam *param,
	const char *stmtName, int resultFormat);

#ifdef __cplusplus
}
#endif
#endif /* !LIBPQTYPES_H */

